/*-------------------------------------------------------------------------+
| Copyright 2011 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.internal;

import static java.util.Collections.emptyList;
import static java.util.Collections.unmodifiableList;
import static org.conqat.lib.commons.collections.CollectionUtils.asUnmodifiable;
import static org.fortiss.tooling.kernel.utils.ExtensionPointUtils.getBundle;
import static org.fortiss.tooling.kernel.utils.ExtensionPointUtils.getConfigurationElements;
import static org.fortiss.tooling.kernel.utils.ExtensionPointUtils.loadClass;
import static org.fortiss.tooling.kernel.utils.LoggingUtils.error;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.conqat.lib.commons.collections.CollectionUtils;
import org.conqat.lib.commons.collections.UnmodifiableList;
import org.eclipse.core.resources.WorkspaceJob;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.ILibrary;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.kernel.ToolingKernelActivator;
import org.fortiss.tooling.kernel.extension.IStorageProvider;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.extension.data.ModelStorageError;
import org.fortiss.tooling.kernel.internal.storage.eclipse.EclipseResourceStorageService;
import org.fortiss.tooling.kernel.introspection.IIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectionItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectiveKernelService;
import org.fortiss.tooling.kernel.introspection.items.PersistencyKISSDetailsItem;
import org.fortiss.tooling.kernel.service.IKernelIntrospectionSystemService;
import org.fortiss.tooling.kernel.service.IPersistencyService;
import org.fortiss.tooling.kernel.service.ITutorialService;
import org.fortiss.tooling.kernel.service.listener.IPersistencyServiceListener;
import org.osgi.framework.Bundle;

/**
 * This class implements the {@link IPersistencyService} interface. All of its
 * public methods are synchronized.
 * 
 * @author hoelzl
 */
public class PersistencyService implements IPersistencyService, IIntrospectiveKernelService {
	/** The singleton instance. */
	private static final PersistencyService INSTANCE = new PersistencyService();

	/** Returns singleton instance of the service. */
	public static PersistencyService getInstance() {
		return INSTANCE;
	}

	/** The prototype provider extension point ID. */
	private static final String EXTENSION_POINT_NAME =
			"org.fortiss.tooling.kernel.modelStorageProvider";

	/** The prototype provider configuration element name. */
	private static final String CONFIGURATION_ELEMENT_NAME = "modelStorageProvider";

	/** Stores the registered storage providers. */
	private final List<IStorageProvider> storageProviderList = new ArrayList<IStorageProvider>();

	/** Stores the top-level element contexts. */
	private final List<ITopLevelElement> elementCache = new LinkedList<ITopLevelElement>();

	/** Stores the dummy top-level element contexts. */
	private Map<EObject, ITopLevelElement> dummyCache = null;

	/** Stores the storage providers and their element caches. */
	private final Map<ITopLevelElement, IStorageProvider> storageProviderCache =
			new HashMap<ITopLevelElement, IStorageProvider>();

	/** Stores the top-level element listeners. */
	private final List<IPersistencyServiceListener> listeners =
			new ArrayList<IPersistencyServiceListener>();

	/** The tutorial service instance. */
	// TODO(VA) Would be of value to explain why the tutorial service is needed in the persistency
	// service...
	private ITutorialService tutorialService;

	/** Initializes the service. */
	public void initializeService() {
		setupStorageProviders();
		for(IStorageProvider p : storageProviderList) {
			if(p instanceof EclipseResourceStorageService) {
				EclipseResourceStorageService ersp = (EclipseResourceStorageService)p;
				ersp.initializeService();
			}
		}
		initializeTopLevelElementContexts();
	}

	/** Starts the service. */
	public void startService() {
		IKernelIntrospectionSystemService.getInstance().registerService(this);

		tutorialService = ITutorialService.getInstance();
		// never call unget, since persistency service never dies
	}

	/** Registers the given storage provider with the service. */
	@Override
	public void registerModelStorageProvider(IStorageProvider provider) {
		storageProviderList.add(provider);
		for(ITopLevelElement context : provider.getTopLevelElements()) {
			elementCache.add(context);
			storageProviderCache.put(context, provider);
			notifyListenersAboutLoad(context);
		}
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionDescription() {
		return getIntrospectionLabel() +
				"\n\nThis service manages the storage of models using a registered persistency mechanism most commonly the local file system storage." +
				"\nIt provides access to top-level elements, which are the entry point to each models command stack." +
				"\nIt also provides a mechanism to add arbitrary EObjects as dummy top-level elements without requiring them " +
				"\nto be backed up by any storage solution (i.e. memory only model). With this the kernel can act as if some model was persisted." +
				"\nThe latter feature is usually used by tests." +
				"\n\nThe service extension point is '" + EXTENSION_POINT_NAME + "'.";
	}

	/** {@inheritDoc} */
	@Override
	public synchronized UnmodifiableList<ITopLevelElement> getTopLevelElements() {
		if(tutorialService != null && tutorialService.isTutorialActive()) {
			EObject activeRootElement = tutorialService.getActiveTutorial().getRootElement();
			return asUnmodifiable(
					elementCache.stream().filter(t -> t.getRootModelElement() == activeRootElement)
							.collect(Collectors.toList()));
		}
		Stream<ITopLevelElement> noDummy =
				elementCache.stream().filter(elt -> !(elt instanceof DummyTopLevelElement));
		return CollectionUtils.asUnmodifiable(noDummy.collect(Collectors.toList()));
	}

	/** {@inheritDoc} */
	@Override
	public synchronized boolean isTopLevelElement(EObject element) {
		for(ITopLevelElement context : elementCache) {
			if(context.getRootModelElement() == element) {
				return true;
			}
		}
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public synchronized boolean isDirty() {
		for(ITopLevelElement context : elementCache) {
			if(context.isDirty()) {
				return true;
			}
		}
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public synchronized void doSave(IProgressMonitor monitor) {
		for(ITopLevelElement context : elementCache) {
			if(context.isDirty()) {
				try {
					EObject rootElement = context.getRootModelElement();
					context.doSave(monitor);
				} catch(CoreException e) {
					error(ToolingKernelActivator.getDefault(), "Error during save operation.", e);
				} catch(IOException e) {
					error(ToolingKernelActivator.getDefault(), "Error during save operation.", e);
				}
			}
		}
	}

	/** {@inheritDoc} */
	@Override
	public synchronized void refreshTopLevelElements(IStorageProvider provider) {
		if(provider == null) {
			return;
		}

		List<ITopLevelElement> providedElements = provider.getTopLevelElements();
		List<ITopLevelElement> removedCacheElements = new LinkedList<ITopLevelElement>();
		// determine cache elements no longer provided
		for(ITopLevelElement top : elementCache) {
			if(provider == storageProviderCache.get(top) && !providedElements.contains(top)) {
				removedCacheElements.add(top);
			}
		}
		// determine and notify added elements
		for(ITopLevelElement top : providedElements) {
			if(!elementCache.contains(top)) {
				elementCache.add(top);
				storageProviderCache.put(top, provider);
				notifyListenersAboutAdd(top);
			}
		}
		// remove elements from cache and notify
		for(ITopLevelElement top : removedCacheElements) {
			elementCache.remove(top);
			storageProviderCache.remove(top);
			notifyListenersAboutRemove(top);
		}
	}

	/** {@inheritDoc} */
	@Override
	public synchronized void addTopLevelElementListener(IPersistencyServiceListener listener) {
		if(!listeners.contains(listener)) {
			listeners.add(listener);
		}
	}

	/** {@inheritDoc} */
	@Override
	public void removeTopLevelElementListener(IPersistencyServiceListener listener) {
		listeners.remove(listener);
	}

	/** Notifies listener about top-level element loading (startup). */
	private synchronized void notifyListenersAboutLoad(ITopLevelElement top) {
		// TODO (#4219) Is this the right way to prevent showing libraries in the navigator
		if(top.getRootModelElement() instanceof ILibrary) {
			return;
		}
		for(IPersistencyServiceListener listener : new ArrayList<>(listeners)) {
			listener.topLevelElementLoaded(top);
		}
	}

	/** Notifies listener about top-level element adding. */
	private synchronized void notifyListenersAboutAdd(ITopLevelElement top) {
		// TODO (#4219) Is this the right way to prevent showing libraries in the navigator
		if(top.getRootModelElement() instanceof ILibrary) {
			return;
		}
		for(IPersistencyServiceListener listener : new ArrayList<>(listeners)) {
			listener.topLevelElementAdded(top);
		}
	}

	/** Notifies listener about top-level element removal. */
	private synchronized void notifyListenersAboutRemove(ITopLevelElement top) {
		// TODO (#4219) Is this the right way to prevent showing libraries in the navigator
		if(top.getRootModelElement() instanceof ILibrary) {
			return;
		}
		for(IPersistencyServiceListener listener : new ArrayList<>(listeners)) {
			listener.topLevelElementRemoved(top);
		}
	}

	/** {@inheritDoc} */
	@Override
	public void notifyTopLevelElementChanged(final ITopLevelElement top) {
		// TODO (#4219) Is this the right way to prevent showing libraries in the navigator
		if(top.getRootModelElement() instanceof ILibrary) {
			return;
		}
		// schedule persistency service notification
		new WorkspaceJob("Notify TopLevelElement Changed") {
			@Override
			public IStatus runInWorkspace(IProgressMonitor monitor) {
				synchronized(PersistencyService.this) {
					for(IPersistencyServiceListener listener : listeners) {
						listener.topLevelElementContentChanged(top);
					}
				}
				return Status.OK_STATUS;
			}
		}.schedule();
	}

	/** {@inheritDoc} */
	@Override
	public synchronized ITopLevelElement getTopLevelElementFor(EObject modelElement) {
		while(modelElement != null) {
			for(ITopLevelElement context : elementCache) {
				if(context.getRootModelElement() == modelElement) {
					return context;
				}
			}
			modelElement = modelElement.eContainer();
		}
		return null;
	}

	/** Initializes the storage provider list from plugin extensions. */
	private void setupStorageProviders() {
		for(IConfigurationElement ce : getConfigurationElements(EXTENSION_POINT_NAME,
				CONFIGURATION_ELEMENT_NAME)) {
			Bundle bundle = getBundle(ce);
			try {
				Class<?> handlerClass = loadClass(ce.getAttribute("provider"), bundle);
				IStorageProvider provider =
						(IStorageProvider)handlerClass.getConstructor().newInstance();
				storageProviderList.add(provider);
			} catch(Exception e) {
				error(ToolingKernelActivator.getDefault(), e.getMessage(), e);
			}
		}
	}

	/** Initializes the top-level element contexts. */
	private void initializeTopLevelElementContexts() {
		for(IStorageProvider provider : storageProviderList) {
			for(ITopLevelElement context : provider.getTopLevelElements()) {
				elementCache.add(context);
				storageProviderCache.put(context, provider);
			}
		}
	}

	/** {@inheritDoc} */
	@Override
	public ITopLevelElement addDummyEObjectAsTopLevelElement(EObject dummyRoot) {
		if(dummyCache == null) {
			dummyCache = new HashMap<EObject, ITopLevelElement>();
		}
		DummyTopLevelElement dummy = new DummyTopLevelElement(dummyRoot);
		dummyCache.put(dummyRoot, dummy);
		elementCache.add(dummy);
		notifyListenersAboutAdd(dummy);
		return dummy;
	}

	/** {@inheritDoc} */
	@Override
	public void removeDummyTopLevelElement(EObject dummy) {
		if(dummyCache == null || !dummyCache.containsKey(dummy)) {
			return;
		}
		ITopLevelElement top = dummyCache.get(dummy);
		elementCache.remove(top);
		dummyCache.remove(dummy);
		notifyListenersAboutRemove(top);
	}

	/** {@inheritDoc} */
	@Override
	public List<ModelStorageError> getAllStorageErrors() {
		List<ModelStorageError> errors = new ArrayList<ModelStorageError>();
		for(IStorageProvider provider : storageProviderList) {
			errors.addAll(provider.getStorageErrors());
		}
		return errors;
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionLabel() {
		return "Persistency Service";
	}

	/** {@inheritDoc} */
	@Override
	public List<IIntrospectionItem> getIntrospectionItems() {
		return emptyList();
	}

	/** {@inheritDoc} */
	@Override
	public boolean showInIntrospectionNavigation() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public IIntrospectionDetailsItem getDetailsItem() {
		return new PersistencyKISSDetailsItem(unmodifiableList(storageProviderList));
	}
}
